/*
 *  game.h
 *  Template
 *
 *  Created by increpare on 28/02/2009.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */


const int tilemap_tw=16;//height in tiles
const int tilemap_th=16;

bool canpass[tilemap_tw*tilemap_th];


const int mapwidth=150;
const int mapheight=150;

int worldmap[mapwidth][mapheight];

const int sectorwidth=50;
const int sectorheight=50;


int px=45;
int py=45;
int pdir=3;
bool moving=false;

/*
 0 home square
 1 brigand
 2 wilderness
 3 black bear
 4 white bear
 5 city
 */

const int mapoutlinecount=8;

int mapoutlines[mapoutlinecount][3][3]=
{

{
{0,2,4},
{1,2,2},
{3,2,5},
},

{
{0,1,3},
{2,2,2},
{4,2,5},
},

{
{4,2,0},
{2,2,1},
{5,2,3},
},

{
{3,1,0},
{2,2,2},
{5,2,4},
},

{
{5,2,4},
{2,2,2},
{3,1,0},
},

{
{5,2,3},
{2,2,2},
{4,1,0},
},

{
{4,2,5},
{1,2,2},
{0,2,3},
},

{
{3,2,5},
{2,2,2},
{0,1,4},
}


};

int mapoutline[3][3] =
{
{0,2,4},
{1,2,2},
{3,2,5},
};


bool showingdialog;
string dialoguestring;


void say(const string& s)
{
	playblip(1);
	showingdialog=true;
	dialoguestring=s;
}


bool hasitem[4];


class NPC{
public:
	int tile;
	int ID;
	int x;
	int y;
	bool dead;
	string text;
	bool brigand;
};

class Entity{
public:	
	int tile;
	int ID;
	int x;
	int y;
	bool dead;
	bool mover;
	string text;
	int motiontimer;
	Entity(int etile,int eid,int ex,int ey,bool edead,bool emover,string etext):
	tile(etile),
	ID(eid),
	x(ex),
	y(ey),
	dead(edead),
	mover(emover),
	text(etext),
	motiontimer(rand()%10)
	{
	};
	void move();
	virtual ~Entity(){};
};


map<int,NPC> npc;
vector<Entity> entities;

class Map{
public:
	int tile[sectorwidth][sectorheight];
};

//s are passive things
//so are conditions that have to be met to continue
enum ScriptKind {
	s_dialogue=0,//say something
	s_remove=1,//remove an npc
	s_give=2,//give the player something
	s_wait=3//wait 'til a condition is met
};

enum WaitKind {
	w_kill=0,//wait 'til you kill someone
	w_talk=1,//wait 'til you talk to someone
	w_get=2//wait 'til you pick up an item (maybe just use talk for this...)
};

class Script{
public:
	ScriptKind kind;
	union {
		struct { char s[100]; } dialoguedata;
		struct { int entity; } removedata;
		struct { int item; } givedata;
		struct { int entity; } interactdata;
		struct { WaitKind waitkind; int target; } waitdata;
	} data;
};

struct Sector {
//these three parameters aren't really all necessary, but for clarity I'm leaving them in
	vector<Map> maps;
};

vector<Script> scripts;
unsigned int scriptindex;
bool finishedscripts=false;

int actors[8];
int curactor;

bool canmoveto(int x, int y)
{
	if (x==0 || y==0||x==(mapwidth-1)||y==(mapheight-1))
		return false;
	
	if (!canpass[worldmap[x][y]])
		return false;
	
	for (int i=0;i<9;i++)
	{
		if (npc[i].x==x && npc[i].y==y)
			return false;
	}
	
	foreach(Entity e,entities)
	{
		if (e.x==x && e.y==y)
			return false;
	}
	return true;
}

bool canmoveentto(int x, int y)
{
	if (x==0 || y==0||x==(mapwidth-1)||y==(mapheight-1))
		return false;
	
	if (!canpass[worldmap[x][y]])
		return false;
	
	//check NPCs first (though there shouldn't really be any such encounters?
	for (int i=0;i<9;i++)
	{
		if (npc[i].x==x && npc[i].y==y)
			return false;
	}
	
	//dont need to worry about self-collisions; only called when x,y different from orig
	foreach(Entity e, entities)
	{
		if (x==e.x && y==e.y)
			return false;
	}
	
	if (x==px && y==py)
		return false;
	
	//should check sector bounds as well maybe?
	
	return true;
}



void Entity::move(){
	if (mover)
	{
		motiontimer=(motiontimer+1)%10;
		
		if (motiontimer==0)
		{
			int tx=x;
			int ty=y;
			switch (rand()%4)
			{
				case 0:ty--;
					break;
				case 1:ty++;
					break;
				case 2:tx--;
					break;
				case 3:tx++;
					break;
			};
			
			if (canmoveentto(tx,ty))
			{
				x=tx;
				y=ty;
			}
		}
		
	}
};

class Mover:public Entity{
public:
	Mover(int etile,int eid,int ex,int ey,bool edead,string etext):
	Entity(etile,eid,ex,ey,edead,1,etext)
	{
	};
};

void do_script()
{	
	if (scriptindex>=scripts.size())
	{
		finishedscripts=true;
		return;
	}
	
	while (true)
	{
		switch (scripts[scriptindex].kind)
		{
			case s_dialogue:
				say(scripts[scriptindex].data.dialoguedata.s);
				scriptindex++;	
				return;
				break;
				
			case s_remove:
				entities[scripts[scriptindex].data.removedata.entity].dead=true;
				scriptindex++;	
				break;
				
			case s_give:
				hasitem[scripts[scriptindex].data.givedata.item]=true;
				scriptindex++;	
				break;
				
			case s_wait:
				return;
				break;
		}	
		
		if (scriptindex>=scripts.size())
		{
			finishedscripts=true;
			return;
		}
		
	}
}
		


void loadscript(string path)
{
	
	lua_State* L;
	/* initialize Lua */
	L = lua_open();
	
	/* load Lua base libraries */
	luaL_openlibs(L);
	
	/* load the script */
	luaL_dofile(L, path.c_str());
	
	
	int i=1;
	scripts.clear();
	scriptindex=0;	
	finishedscripts=false;

	
	lua_getglobal(L,"scripts");
	while(true)
	{
		Script scr;
		lua_pushinteger(L,i);
		
		//look at first element (tag)
		//call table
		lua_gettable(L, -2 );
		if (lua_isnil(L,-1))
		{
			break;
		}
		
		
		lua_pushinteger(L,1);
		lua_gettable(L,-2);
		string command = lua_tostring(L,-1);
		lua_pop(L,1);
		
		if (command=="say")
		{
			scr.kind=s_dialogue;
			
			lua_pushinteger(L,2);
			lua_gettable(L,-2);
			strcpy( scr.data.dialoguedata.s,lua_tostring(L,-1));
			lua_pop(L,1);
			
		}
		else if (command=="remove")
		{
			scr.kind=s_remove;
			
			lua_pushinteger(L,2);
			lua_gettable(L,-2);
			scr.data.removedata.entity=lua_tointeger(L,-1);
			lua_pop(L,1);
		}
		else if (command=="give")
		{			
			scr.kind=s_give;
			
			lua_pushinteger(L,2);
			lua_gettable(L,-2);
			scr.data.givedata.item=lua_tointeger(L,-1);
			lua_pop(L,1);					
		}
		else if (command=="wait")
		{
			scr.kind=s_wait;
			
			lua_pushinteger(L,2);
			lua_gettable(L,-2);
			string t=lua_tostring(L,-1);
			lua_pop(L,1);
			
			if (t=="kill")
			{
				scr.data.waitdata.waitkind=w_kill;
			}
			else if (t=="talk")
			{
				scr.data.waitdata.waitkind=w_talk;
			}
			else if (t=="get")
			{
				scr.data.waitdata.waitkind=w_get;
			}
			else 
			{
				std::cerr<<"ERROR LOADING SCRIPT, SCRIPT LINE "<<i<<endl;
			}		
			
			lua_pushinteger(L,3);
			lua_gettable(L,-2);			
			scr.data.waitdata.target=lua_tointeger(L,-1);
			lua_pop(L,1);
			
		}
		else
		{
			std::cerr<<"ERROR LOADING SCRIPT, SCRIPT LINE "<<i<<endl;
		}
		
		lua_pop(L,1);//pop the individual script line
		
		scripts.push_back(scr);
		i++;
	}
	lua_pop(L,1);
	
	
	
	
	
	lua_close(L);
}

Sector loadsector(string path)
{
	Sector s;
		
	lua_State* L;
	/* initialize Lua */
	L = lua_open();
	
	/* load Lua base libraries */
	luaL_openlibs(L);
	
	/* load the script */
	luaL_dofile(L, path.c_str());
	
	//LOAD MAP
	
	s.maps.clear();
	int i=1;
	lua_getglobal(L,"maps");
	while(true)
	{
		Map m;
		lua_pushinteger(L,i);
		
		
		//call table
		lua_gettable(L, -2 );
		if (lua_isnil(L,-1))
		{
			break;
		}
		
		for (int j=1;j<=sectorheight;j++)
		{
			//call row
			
			lua_pushinteger(L,j);
			lua_gettable(L,-2);
			
			for (int i=1;i<=sectorwidth;i++)
			{
				lua_pushinteger(L,i);
				lua_gettable(L,-2);
				m.tile[i-1][j-1]=lua_tointeger(L,-1);
				lua_pop(L,1);
			}
			
			lua_pop(L,1);
		}
		
		lua_pop(L,1);
		
		s.maps.push_back(m);
		i++;
	}
	lua_pop(L,1);
	

	lua_close(L);

	return s;
}

//tile coord to int
int tn(int x, int y)
{		
	return x+y*tilemap_tw;
}

int freearound(int x, int y)
{
	return (
			canpass[worldmap[x-1][y-1]] && canpass[worldmap[x][y-1]] && canpass[worldmap[x+1][y-1]] && 
			canpass[worldmap[x-1][y  ]] && canpass[worldmap[x][y  ]] && canpass[worldmap[x+1][y  ]] && 
			canpass[worldmap[x-1][y+1]] && canpass[worldmap[x][y+1]] && canpass[worldmap[x+1][y+1]] 
			);
}

bool generictile(int t)
{

	

	return (t==tn(13,12) || 
			t==tn(14,12) ||
			t==tn(15,12) ||
			t==tn(13,13) ||
			t==tn(14,13));
}

void addentity(int x, int y)
{
	//no npc entities in towns
	if (mapoutline[x/sectorwidth][y/sectorheight]==0 || mapoutline[x/sectorwidth][y/sectorheight]==5)
		return;
	
	if (worldmap[x][y]==tn(13,12))//GRASS
	{
		entities.push_back(Mover(tn(12+rand()%4,5),0,x,y,0,""));
	}
	else if (worldmap[x][y]==tn(15,12))//swamp
	{
		entities.push_back(Mover(tn(14+rand()%2,8),0,x,y,0,""));	
	}
	else if (worldmap[x][y]==tn(14,13))//desert
	{
		entities.push_back(Mover(tn(14+rand()%2,9),0,x,y,0,""));		
	}
	else if (worldmap[x][y]==tn(13,13))//ice
	{
		entities.push_back(Mover(tn(12+rand()%2,10),0,x,y,0,""));		
	}
}

void addobstacle(int x, int y)
{
	if (worldmap[x][y]==tn(13,12))//GRASS
	{
		switch (rand()%3)
		{
			case 0:
				worldmap[x][y]=tn(4,0);
				break;
			case 1:
				worldmap[x][y]=tn(11,5);				
				break;
			case 2:
				worldmap[x][y]=tn(11,6);				
				break;
		}
	}
	else if (worldmap[x][y]==tn(15,12))//swamp
	{
		switch(rand()%5)
		{
			case 0:
				worldmap[x][y]=tn(1,5);
				break;
			case 1:
				worldmap[x][y]=tn(2,5);
				break;
			case 2:
				worldmap[x][y]=tn(1,6);
				break;
			case 3:
				worldmap[x][y]=tn(2,6);
				break;
			case 4:
				worldmap[x][y]=tn(11,6);
				break;
		}
	}
	else if (worldmap[x][y]==tn(14,13))//desert
	{
		switch(rand()%1)
		{
			case 0:
				worldmap[x][y]=tn(9,7);
				break;
		}		
	}
	
	
	//water should never find its way into this function
	//ice has no obstacles at the moment...should add some?
}

void prettify(int x, int y)
{
	if (worldmap[x][y]==tn(13,12))//GRASS
	{
		switch (rand()%4)
		{
			case 0:
				worldmap[x][y]=tn(1,1);
				break;
			case 1:
				worldmap[x][y]=tn(1,2);				
				break;
			case 2:
				worldmap[x][y]=tn(2,1);				
				break;
			case 3:
				worldmap[x][y]=tn(2,2);				
				break;
		}
	}
	else if (worldmap[x][y]==tn(15,12))//swamp
	{
		switch(rand()%4)
		{
			case 0:
				worldmap[x][y]=tn(5,5);
				break;
			case 1:
				worldmap[x][y]=tn(5,6);
				break;
			case 2:
				worldmap[x][y]=tn(6,5);
				break;
			case 3:
				worldmap[x][y]=tn(6,6);
				break;
		}
	}
	else if (worldmap[x][y]==tn(14,13))//desert
	{
		switch(rand()%11)
		{
			case 0:
				worldmap[x][y]=tn(8,11);
				break;
			case 1:
				worldmap[x][y]=tn(9,11);
				break;
			case 2:
				worldmap[x][y]=tn(8,12);
				break;
			case 3:
				worldmap[x][y]=tn(5,2);
				break;
			case 4:
				worldmap[x][y]=tn(8,1);
				break;
			case 5:
				worldmap[x][y]=tn(8,2);
				break;
			case 6:
				worldmap[x][y]=tn(9,1);
				break;
			case 7:
				worldmap[x][y]=tn(9,2);
				break;
			case 8:
				worldmap[x][y]=tn(8,11);
				break;
			case 9:
				worldmap[x][y]=tn(9,11);
				break;
			case 10:
				worldmap[x][y]=tn(8,12);
				break;
		}		
	}
	else if (worldmap[x][y]==tn(13,13))//ice
	{
		switch(rand()%4)
		{
			case 0:
				worldmap[x][y]=tn(1,9);
				break;
			case 1:
				worldmap[x][y]=tn(2,9);
				break;
			case 2:
				worldmap[x][y]=tn(1,10);
				break;
			case 3:
				worldmap[x][y]=tn(2,10);
				break;
		}			 
	}
	else if (worldmap[x][y]==tn(14,12))//water
	{
		switch(rand()%4)
		{
			case 0:
				worldmap[x][y]=tn(8,9);
				break;
			case 1:
				worldmap[x][y]=tn(8,10);
				break;
			case 2:
				worldmap[x][y]=tn(9,9);
				break;
			case 3:
				worldmap[x][y]=tn(9,10);
				break;
		}					
	}
	
}

int hometowncount;
int foreigncount;
/*
 hometown
 0:chief 
 1:person
 2:person
 3:brigand
 4:black bear
 5:white bear
 6:person
 7:person
 8:person
 */

char* dialogues[10]={
"it is such a lovely day",
"i have a cake in the over",
"this is such a lovely place",
"it is dangerous outside this village",
"i am wary of everyone",
"you look so sad",
"i would like to see you smile",
"you cannot stay sad forever",
"you look so lonely",
"i hope you find her"
};
void processtileforentity(int x, int y)
{
	if (worldmap[x][y]==tn(13,14))//player start
	{
		if (mapoutline[x/sectorwidth][y/sectorheight]==0)//if in home town
		{
			px=x;
			py=y;
			pdir=2;
		}
		worldmap[x][y]=tn(13,12);
	}
	else if (worldmap[x][y]==tn(14,14))//entity
	{
		if (mapoutline[x/sectorwidth][y/sectorheight]==0)//if in home town
		{
			if (actors[curactor]<4)
				npc[hometowncount]=((NPC){tn(12+actors[curactor],6),0,x,y,0,dialogues[rand()%10],0});		
			else
				npc[hometowncount]=((NPC){tn(12+actors[curactor]-4,7),0,x,y,0,dialogues[rand()%10],0});					
				
			worldmap[x][y]=tn(13,12);
			curactor++;
			hometowncount++;			
		}
		else
		{
			if (actors[curactor]<4)
				npc[6+foreigncount]=(NPC){tn(12+actors[curactor],6),0,x,y,0,dialogues[rand()%10]};		
			else
				npc[6+foreigncount]=(NPC){tn(12+actors[curactor]-4,7),0,x,y,0,dialogues[rand()%10]};					
			worldmap[x][y]=tn(13,12);
			curactor++;	
			foreigncount++;			
		}
	}
	else if (worldmap[x][y]==tn(13,15))//BEAR
	{
		if (mapoutline[x/sectorwidth][y/sectorheight]==3)//if first bear
			npc[4]=((NPC){tn(12+rand()%2,8),0,x,y,0,"you have no business with the black bear",0});		
		else
			npc[5]=((NPC){tn(12+rand()%2,9),0,x,y,0,"you have no business with the white bear",0});		
		
		worldmap[x][y]=worldmap[x+1][y];		
	}
	else if (worldmap[x][y]==tn(15,14))//BRIGAND
	{
		if (actors[curactor]<4)
			npc[3]=((NPC){tn(12+actors[curactor],6),0,x,y,0,"get away from here",0});		
		else
			npc[3]=((NPC){tn(13+actors[curactor]-4,6),0,x,y,0,"get away from here",0});					
		
		worldmap[x][y]=worldmap[x][y+1];
		curactor++;
		
	}
		
}

//replaces generic tiles with specific ones; introduces obstacles
void process_map()
{
	//1: place obstacles
	for (int i=1;i<mapwidth-1;i++)
	{
		for (int j=1;j<mapwidth-1;j++)
		{
			if (rand()%10==0)
				{
					if (generictile(worldmap[i][j]) && freearound(i,j))
					{
						addobstacle(i,j);
					}
				}
		}
	}
	
	//1: add entities
	for (int i=1;i<mapwidth-1;i++)
	{
		for (int j=1;j<mapwidth-1;j++)
		{
			if (rand()%10==0)
			{
				if (generictile(worldmap[i][j])&& freearound(i,j))
				{
					addentity(i,j);
				}
			}
		}
	}
	
	//2:look for players/add npcs
	
	hometowncount=0;
	foreigncount=0;
	for (int i=0;i<mapwidth;i++)
	{
		for (int j=0;j<mapwidth;j++)
		{
			processtileforentity(i,j);
		}
		
	}
	
	//3: prettify remaining generic tiles
	for (int i=0;i<mapwidth;i++)
	{
		for (int j=0;j<mapwidth;j++)
		{
			if (generictile(worldmap[i][j]))
			{
				prettify(i,j);
			}
		}
	}	
}

void drawsector(const Sector& s,int sx, int sy)
{
	int m =rand()%(s.maps.size());
	for (int i=0;i<sectorwidth;i++)
	{
		for (int j=0;j<sectorheight;j++)
		{
			worldmap[sx*sectorwidth+i][sy*sectorheight+j]=s.maps[m].tile[i][j]-1;
		}
	}	
}

void initmap()
{
	
	entities.clear();
	npc.clear();
	//do actor stuff
	curactor=0;
	
	for (int i=0;i<8;i++)
	{
		actors[i]=i;
	}
	
	//knuth random perm
	for (int i=0;i<8-1;i++)
	{
		int j;
		
		j=rand()%(i+1);
		swap(actors[i],actors[j]);
		
	}
	
	
	//other
	
	moving=false;
	
	vector<Sector> sectors;
	
	
	for (int i=0;i<mapwidth;i++)
	{
		for (int j=0;j<mapheight;j++)
		{
			worldmap[i][j]=rand()%16*8;
		}
	}
	
	showingdialog=false;
	
	//TOP-LEFT and BOTTOM-RIGHT SECTORS FIRST
	loadscript("lua/script.lua");

	sectors.push_back(loadsector("lua/town.lua"));
	sectors.push_back(loadsector("lua/camp.lua"));
	sectors.push_back(loadsector("lua/wilderness.lua"));
	sectors.push_back(loadsector("lua/bear.lua"));

	int chosenoutline=rand()%mapoutlinecount;
	
	for (int i=0;i<3;i++)
		for (int j=0;j<3;j++)
		{
			mapoutline[i][j]=mapoutlines[chosenoutline][i][j];
		}
	for (int i=0;i<3;i++)
		for (int j=0;j<3;j++)
		{
			switch (mapoutline[i][j])
			{
				case 0://towns
				case 5:					
					drawsector(sectors[0],i,j);								
					break;
				case 1://brigand
					drawsector(sectors[1],i,j);								
					break;
				case 2://wilderness
					drawsector(sectors[2],i,j);													
					break;
				case 3://bear places
				case 4:
					drawsector(sectors[3],i,j);													
					break;
					break;
				
			}
		}

	
	process_map();


	do_script();

}
